# OpenAI chat made by Kevin Gauthun, 7-22-2023
# uses openAI api, allows selection from any available model you have API 
# access to. To use gpt-4 you must have an API key for gpt-4 available. 
# keys are registered under
# https://platform.openai.com/account/api-keys
# once the program runs, hit ok to enter a new alias, then enter your API key
# then enter the password to encrypt your key in base64
# it may ask for you to re enter your password
# on second boot of the program you can select your previously defined alias
# enter the password you set previously (even if it says "enter new password to encrypt api key")
# This is actually asking for the password to read the hash encoded API key
# i will be fixing this later so the entry forms are not as redundant/confusing.
# note: content/context is the context of the chat. An example would be:
# context: You are a junior


import os
import json
import base64
import openai
import tkinter as tk
from tkinter import ttk
from tkinter import simpledialog
from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
import re 

# Function to derive a key from a password
def derive_key(password, salt):
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        iterations=100000,
    )
    return base64.urlsafe_b64encode(kdf.derive(password))

# Function to check if API key is valid
def check_api_key(api_key):
    openai.api_key = api_key
    try:
        # Try to list the available models
        openai.Model.list()
        return True
    except openai.OpenAIError:
        return False
    
def get_alias(root):
    alias = simpledialog.askstring("Input", "Please enter your alias:", parent=root)
    return alias

class APIKeyDialog(simpledialog.Dialog):
    def __init__(self, parent, api_keys_dict):
        self.api_keys_dict = api_keys_dict
        super().__init__(parent)
        
    def body(self, parent):
        tk.Label(parent, text="Please select or enter your alias:").grid(row=0)
        self.api_key_entry = ttk.Combobox(parent)
        self.api_key_entry.grid(row=1, padx=5, pady=5)
        stored_keys = list(self.api_keys_dict.keys())
        stored_keys.append("Enter a new alias")
        self.api_key_entry['values'] = stored_keys
        self.api_key_entry.current(0)
        return self.api_key_entry

    def validate(self):
        alias = self.api_key_entry.get()
        if alias == "Enter a new alias":
            alias = get_alias(self.parent)
            api_key = simpledialog.askstring(
                "API Key", 
                "Please enter your new OpenAI API key:", 
                parent=self.parent, 
                show='*'  # This will hide the input
            )
            if check_api_key(api_key):
                self.result = (alias, api_key)
                return True
            else:
                tk.messagebox.showerror("Error", "The API key you entered is invalid. Please try again.")
                return False
        else:
            self.result = (alias, self.api_keys_dict[alias])
            return True

def is_base64(s):
    if len(s) % 4 > 0:  # base64 string lengths are multiples of 4
        return False
    if re.match('^[A-Za-z0-9+/]+[=]{0,2}$', s) is not None:  # check the content
        return True
    return False

def get_password(root):
    password = simpledialog.askstring("Password", "Please enter a password to secure your API key:", parent=root, show='*')
    return password.encode("utf-8")  # Convert the password to bytes

def get_api_key(root, api_keys_dict):
    global alias, api_key
    dialog = APIKeyDialog(root, api_keys_dict)
    alias, api_key = dialog.result

    # Check if the api_key is already encrypted
    if isinstance(api_key, str) and not is_base64(api_key):
        # Get a password from the user
        password = get_password(root)

        # Generate a random salt
        salt = os.urandom(16)

        # Derive a key from the password
        key = derive_key(password, salt)

        # Create a cipher suite
        cipher_suite = Fernet(key)

        # Encrypt the API key
        cipher_text = cipher_suite.encrypt(api_key.encode())  # Convert API key to bytes before encryption

        # Store the encrypted API key in the dictionary using the alias as a reference
        api_keys_dict[alias] = {
            'key': cipher_text.decode(),
            'salt': base64.urlsafe_b64encode(salt).decode()
        }

        # Save the dictionary of API keys to a file
        with open(api_keys_file, "w") as file:
            json.dump(api_keys_dict, file)
            
# Initialize tkinter
root = tk.Tk()
root.withdraw()  # Hide the main window

# Define the API key file path
api_keys_file = "api_keys.json"

# Load existing API keys if the file exists
if os.path.exists(api_keys_file):
    with open(api_keys_file, 'r') as file:
        api_keys_dict = json.load(file)
else:
    api_keys_dict = {}

# Declare alias and api_key as global variables
alias = None
api_key = None

# Get the alias and API key from the user
get_api_key(root, api_keys_dict)

# Get a password from the user
password = get_password(root)

# Get the salt
salt = base64.urlsafe_b64decode(api_keys_dict[alias]['salt'])

# Derive a key from the password
key = derive_key(password, salt)

# Create a cipher suite
cipher_suite = Fernet(key)

# Decrypt the API key if it's a dictionary (which means it's already encrypted)
if isinstance(api_key, dict):
    encrypted_api_key = api_key['key']
    decrypted_api_key = cipher_suite.decrypt(encrypted_api_key.encode())
    openai.api_key = decrypted_api_key.decode()

# Encrypt the API key if it's a string and not a base64 encoded string already
elif isinstance(api_key, str) and not is_base64(api_key):
    cipher_text = cipher_suite.encrypt(api_key.encode())  
    

    # Store the encrypted API key in the dictionary using the alias as a reference
    api_keys_dict[alias] = {
        'key': cipher_text.decode(),
        'salt': base64.urlsafe_b64encode(salt).decode()
    }

    # Save the dictionary of API keys to a file
    with open(api_keys_file, "w") as file:
        json.dump(api_keys_dict, file)

tk.messagebox.showinfo("Success", "API key stored successfully.")

root.deiconify()

root.configure(bg='lightblue')

# Create a frame for the model selection
model_frame = tk.Frame(root, bg='lightblue')
model_frame.pack(fill='x')

# Create variable for model selection
model_var = tk.StringVar(root)
model_var.set("gpt-3.5-turbo-16k") # default value

# Create dropdown menu for model selection
model_dropdown = tk.OptionMenu(model_frame, model_var, "gpt-3.5-turbo-16k", "gpt-3.5-turbo", "gpt-3.5-turbo-16k-0613", "gpt-4-32k", "gpt-4")
model_dropdown.pack(side='right')

# Create label for current model
model_label = tk.Label(model_frame, text="Current Model:", bg='lightblue', fg='black')
model_label.pack(side='right')

# Create a frame for the chat log and scrollbar
chat_frame = tk.Frame(root, bg='lightblue')
chat_frame.pack(fill='both', expand=True)

# Create chat log
chat_log = tk.Text(chat_frame, state='disabled', bg='white', fg='black', bd=2, relief='solid')
chat_log.tag_configure('user', foreground='blue')
chat_log.tag_configure('bot', foreground='royalblue')
# Create scrollbar
scrollbar = tk.Scrollbar(chat_frame, command=chat_log.yview)

# Link scrollbar and chat log
chat_log['yscrollcommand'] = scrollbar.set

# Pack chat_log and scrollbar
chat_log.pack(side='left', fill='both', expand=True)
scrollbar.pack(side='right', fill='y')

# Create entry field
entry_field = tk.Entry(root, bg='white', fg='black', bd=2, relief='solid')
entry_field.pack(fill='x')

# Create a frame for temperature selection
temp_frame = tk.Frame(root, bg='lightblue')
temp_frame.pack(fill='x')

# Create variable for temperature
temp_var = tk.DoubleVar(root)
temp_var.set(0.5) # default value

# Create label for temperature
temp_label = tk.Label(temp_frame, text="Temperature:", bg='lightblue', fg='black')
temp_label.pack(side='left')

# Create slider for temperature
temp_slider = tk.Scale(temp_frame, from_=0, to=1, orient='horizontal', resolution=0.01, length=400, variable=temp_var)
temp_slider.pack(side='right')

# Create a frame for context entry
context_frame = tk.Frame(root, bg='lightblue')
context_frame.pack(fill='x')

# Create label for context
context_label = tk.Label(context_frame, text="Context:", bg='lightblue', fg='black')
context_label.pack(side='left')

# Create entry field for context
context_entry = tk.Entry(context_frame, bg='white', fg='black', bd=2, relief='solid')
context_entry.pack(fill='x')

def send_message(event=None):
    message = entry_field.get().strip()
    if message:
        chat_log.configure(state='normal')

        # Insert the user's text with the 'user' tag
        chat_log.insert('end', 'You: ' + message + '\n', 'user')

        # Add bubble effect
        chat_log.tag_configure('user', background='lightblue', lmargin1='5m', lmargin2='5m', rmargin='5m')

        chat_log.configure(state='disabled')

        # Get temperature and context
        temperature = temp_var.get()
        context = context_entry.get().strip()

        # Call OpenAI API
        model_name = model_var.get()
        response = openai.ChatCompletion.create(
            model=model_name,
            messages=[
                {"role": "system", "content": context or "You are a helpful assistant."},
                {"role": "user", "content": message}
            ],
            temperature=temperature
        )

        chat_log.configure(state='normal')

        # Insert the bot's text with the 'bot' tag
        chat_log.insert('end', model_name + ': ' + response['choices'][0]['message']['content'] + '\n', 'bot')

        # Add bubble effect
        chat_log.tag_configure('bot', background='lightgreen', lmargin1='5m', lmargin2='5m', rmargin='5m')

        chat_log.configure(state='disabled')
        chat_log.see('end')  # Auto-scroll to the end
        

# Create send button
send_button = tk.Button(root, text='Send', command=send_message, bg='lightpink', fg='black', bd=2, relief='raised')
send_button.pack(fill='x')

# Bind Enter key to send message
root.bind('<Return>', send_message)

root.mainloop()